"""
Tasks that can be run from a Toyz web application
"""
# Copyright 2015 by Fred Moolekamp
# License: LGPLv3
from __future__ import division,print_function
import os

from toyz.utils import core
from toyz.utils.errors import ToyzJobError
import astrotoyz as astro
from toyz.web import session_vars

def get_img_info(toyz_settings, tid, params):
    """
    Get image info specific to astronomy FITS images (like physical coords, wcs, etc.)
    """
    core.check4keys(params, ['file_info', 'img_info'])
    response = {
        'id': 'get_img_info',
        'img_info': astro.viewer.get_img_info(**params)
    }
    return response

def get_img_data(toyz_settings, tid, params):
    """
    Get data from an image or FITS file
    """
    core.check4keys(params, ['data_type', 'file_info', 'img_info'])
    response = astro.viewer.get_img_data(**params)
    return response

def get_2d_fit(toyz_settings, tid, params):
    """
    Get desired fit for 2d data array
    """
    import numpy as np
    core.check4keys(params, ['file_info', 'fit_type', 'x', 'y', 'width', 'height'])
    response = astro.viewer.get_2d_fit(**params)
    return response

def create_catalog(toyz_settings, tid, params):
    """
    Save a catalog
    """
    core.check4keys(params, ['settings', 'cid'])
    catalog = astro.catalog.get_catalog(create='yes', **params)
    # Store the catalog in the session for future use
    session_vars.catalogs[catalog.cid] = catalog
    # Translate the dataframe to a format that can be JSON encoded
    # (this includes translating NaN values to strings that will be
    # translated back to NaN in the client)
    index = list(catalog.index.names)
    dataframe = {
        'columns': index+catalog.columns.values.tolist(),
        'index': index,
        'data': catalog.astype(object).fillna('NaN').values.tolist()
    }
    cat_info = {
        'cid': catalog.cid,
        'name': catalog.name,
        'settings': catalog.settings,
        'dataframe': dataframe
    }
    response = {
        'id': 'create_catalog',
        'cat_info': cat_info
    }
    return response

def save_catalog(toyz_settings, tid, params):
    """
    Save a catalog
    """
    core.check4keys(params, ['cid'])
    catalog = astro.catalog.get_catalog(params['cid'], create='fail')
    if 'filepath' in params:
        catalog.save(params['filepath'])
    else:
        catalog.save()
    response = {
        'id': 'save_catalog',
        'status': 'success'
    }
    return response

def select_src(toyz_settings, tid, params):
    """
    Get the nearest source to a given point
    """
    core.check4keys(params, ['cid'])
    catalog = astro.catalog.get_catalog(params['cid'], create='fail')
    src = catalog.select_src(**params)
    if len(src)>0:
        status = 'success'
    else:
        status = 'failed: no source found'
    response = {
        'id': 'select_src',
        'status': status,
        'src': src
    }
    return response

def add_src(toyz_settings, tid, params):
    """
    Add a source to the catalog
    """
    core.check4keys(params, ['cid', 'file_info', 'src_info'])
    catalog = astro.catalog.get_catalog(params['cid'], create='fail')
    del params['cid']
    src = catalog.add_src(**params)
    response = {
        'id': 'add_src',
        'src': src
    }
    if len(src)>0:
        response['status'] = 'success'
    else:
        response['status'] = 'failed: source already exists'
    #print('catalog after added source')
    #print(catalog)
    return response

def delete_src(toyz_settings, tid, params):
    """
    Delete a source from the catalog
    """
    core.check4keys(params, ['cid', 'src_info'])
    catalog = astro.catalog.get_catalog(params['cid'], create='fail')
    status = catalog.delete_src(params['src_info'])
    response = {'id': 'delete_src'}
    if status is True:
        response['status'] = 'success'
    else:
        response['status'] = 'failed'
    return response

def detect_sources(toyz_settings, tid, params):
    """
    Detect sources and create an object catalog in the current session
    """
    core.check4keys(params,['file_info', 'cid', 'settings'])
    catalog = astro.catalog.detect_sources(**params)
    #print('catalog', catalog)
    #print('catlog shape', catalog.shape)
    print('generating response')
    response = {
        'id': 'detect_sources',
        'sources': catalog.get_markers(),
        'settings': catalog.settings
    }
    print('response', response)
    return response

def wcs2px(toyz_Settings, tid, params):
    """
    Align all images in the viewer with the world coordinates of the current image
    """
    import toyz.web
    from astrotoyz.viewer import get_wcs
    import numpy as np
    print('filepath:', params['file_info']['filepath'],'\n\n')
    core.check4keys(params, ['file_info', 'img_info', 'ra', 'dec'])
    hdulist = toyz.web.viewer.get_file(params['file_info'])
    wcs = get_wcs(params['file_info'], hdulist)
    if wcs is None:
        raise astrotoyz.core.AstroToyzError('Unable to load WCS for the current image')
    data = hdulist[int(params['file_info']['frame'])].data
    ra = params['ra']
    dec = params['dec']
    wcs_array = wcs.wcs_world2pix(np.array([[ra, dec]]), 1)
    response = {
        'id': 'wcs2px',
        'x': int(round(wcs_array[0][0])),
        'y': int(round(wcs_array[0][1])),
        # next two lines for testing
        'ra': ra,
        'dec': dec
    }
    print('px coords response', response)
    return response

def load_sextractor(toyz_settings, tid, params):
    """
    Load sextractor configuration and parameters
    """
    import astrotoyz.astromatic.sex as sex
    params, param_order = sex.build_param_gui()
    response = {
        'id': 'load_sextractor',
        'params': params,
        'param_order': param_order
    }
    return response

def run_sextractor(toyz_settings, tid, params):
    """
    Run sextractor using parameters defined in the client
    """
    import astrotoyz.astromatic.sex as sex
    import toyz.utils.db
    import shutil
    
    core.check4keys(params, ['config', 'params', 'filename'])
    shortcuts = toyz.utils.db.get_param(toyz_settings.db, 'shortcuts', user_id=tid['user_id'])
    # Just in case the user did something dumb, like remove his/her temp path
    if 'temp' not in shortcuts:
        raise ToyJobError("You removed your 'temp' shortcut, DON'T DO THAT!" +
            "Refresh your browser and it will be restored")
    # Create a path for temporary files that will be erased once this has been completed
    temp_path = os.path.join(shortcuts['temp'], tid['session_id'], str(tid['request_id']))
    core.create_paths(temp_path)
    # Run SExtractor
    params['temp_path'] = temp_path
    result = sex.run_sextractor(**params)
    # Remove the temporary path form the server
    #shutil.rmtree(temp_path)
    
    if result!='success':
        response = {
            'id': 'ERROR',
            'error': result
        }
    else:
        response = {
            'id': 'run_sextractor',
            'result': result
        }
    return response


"""
IO tools for Astro-Toyz. This is basically a wrapper for astropy's io module to read
and write tables
"""
# Copyright 2015 by Fred Moolekamp
# License: LGPLv3

from toyz.utils.core import merge_dict

# TODO: Complete io_module for all possible astropy file io types
io_modules = {
    'astropy': {
        'ascii': {
            'load': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                    'format': {
                        'lbl': 'format',
                        'type': 'select',
                        'options': [
                            'aastex',
                            'ascii',
                            'ascii.aastex',
                            'ascii.basic',
                            'ascii.cds',
                            'ascii.commented_header',
                            'ascii.daophot',
                            'ascii.ecsv',
                            'ascii.fixed_width',
                            'ascii.fixed_width_no_header',
                            'ascii.fixed_width_two_line',
                            'ascii.html',
                            'ascii.ipac',
                            'ascii.latex',
                            'ascii.no_header',
                            'ascii.rdb',
                            'ascii.sextractor',
                            'ascii.tab',
                            'ascii.csv',
                        ]
                    }
                },
                'optional': {
                    'delimiter': {'lbl': 'delimiter'}
                    # TODO Complete optional parameters for ascii type
                }
            },
            'save': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                    'format': {
                        'lbl': 'format',
                        'type': 'select',
                        'options': [
                            'aastex',
                            'ascii',
                            'ascii.aastex',
                            'ascii.basic',
                            'ascii.cds',
                            'ascii.commented_header',
                            'ascii.daophot',
                            'ascii.ecsv',
                            'ascii.fixed_width',
                            'ascii.fixed_width_no_header',
                            'ascii.fixed_width_two_line',
                            'ascii.html',
                            'ascii.ipac',
                            'ascii.latex',
                            'ascii.no_header',
                            'ascii.rdb',
                            'ascii.sextractor',
                            'ascii.tab',
                            'ascii.csv',
                        ]
                    }
                }
            },
            'load2save': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'save2load': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'load_fn': 'astropy_read',
            'save_fn': 'astropy_write'
        },
        'cds': {
            'load': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'save': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'load2save': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'save2load': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'load_fn': 'astropy_read',
            'save_fn': 'astropy_write'
        },
        'daophot': {
            'load': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'save': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'load2save': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'save2load': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'load_fn': 'astropy_read',
            'save_fn': 'astropy_write'
        },
        'fits': {
            'load': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'save': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'load2save': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'save2load': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'load_fn': 'astropy_read',
            'save_fn': 'astropy_write'
        },
        'hdf5': {
            'load': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                    'path': {'lbl': 'path'}
                }
            },
            'save': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                    'path': {'lbl': 'path'},
                    'append': {
                        'lbl': 'append',
                        'prop': {
                            'type': 'checkbox',
                            'checked': False
                        }
                    },
                    'overwrite': {
                        'lbl': 'append',
                        'prop': {
                            'type': 'checkbox',
                            'checked': False
                        }
                    }
                }
            },
            'load2save': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'save2load': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'load_fn': 'astropy_read',
            'save_fn': 'astropy_write'
        },
        'html': {
            'load': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'save': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'load2save': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'save2load': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'load_fn': 'astropy_read',
            'save_fn': 'astropy_write'
        },
        'ipac': {
            'load': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'save': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'load2save': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'save2load': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'load_fn': 'astropy_read',
            'save_fn': 'astropy_write'
        },
        'latex': {
            'load': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'save': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'load2save': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'save2load': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'load_fn': 'astropy_read',
            'save_fn': 'astropy_write'
        },
        'rdb': {
            'load': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'save': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'load2save': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'save2load': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'load_fn': 'astropy_read',
            'save_fn': 'astropy_write'
        },
        'votable': {
            'load': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'save': {
                'type': 'div',
                'params': {
                    'filename': {
                        'lbl': 'filename',
                        'file_dialog': True
                    },
                }
            },
            'load2save': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'save2load': {
                'remove': [],
                'warn': {},
                'convert': {}
            },
            'load_fn': 'astropy_read',
            'save_fn': 'astropy_write'
        }
    }
}

def astropy_read(file_type, **file_options):
    """
    Read into an astropy table
    """
    from astropy.table import Table
    filename = file_options['filename']
    options = merge_dict({}, file_options)
    del options['filename']
    # Make sure to use the proper format
    if file_type!='ascii':
        options['format'] = file_type
    data = Table.read(filename, **options)
    return data

def astropy_write(data, file_type, **file_options):
    """
    Write an astropy table to a file
    """
    from astropy.table import Table
    filename = file_options['filename']
    options = merge_dict({}, file_options)
    del options['filename']
    # Make sure to use the proper format
    if file_type!='ascii':
        options['format'] = file_type
    data.write(filename, **options)


"""
Configuration file required for all new Toyz
"""
from __future__ import division,print_function
import os

import astrotoyz.io

##################
# Custom Methods #
##################

# Root directory of the config.py file
root = os.path.abspath(os.path.join(os.path.dirname(__file__)))

####################
# Required methods #
####################

# List of static and template paths
static_paths = [os.path.join(root, 'static')]
template_paths = [os.path.join(root, 'templates')]

# Tiles added by your toy
workspace_tiles = {
    'astro_viewer': {
        'name': 'Astro Viewer', # Name of the tile to show up in the workspace tiles list
        'namespace': 'Toyz.Astro.Viewer', # Namespace of the template (to check for dependencies)
        'url': '/toyz/static/astrotoyz/viewer.js' # Javascript file to load
    }
}

# Urls to add to the 'Toyz' tab on the home page. The keys are the text that will appear and the
# values are the urls
toyz_urls = {
}

# If any tornado.RequestHandler templates have parameters, special functions to render
# the page must be defined here
render_functions = {
}

# Custom IO modules
io_modules = astrotoyz.io.io_modules
load_functions = {
    'astropy_read': astrotoyz.io.astropy_read
}
save_functions = {
    'astropy_write': astrotoyz.io.astropy_write
}

# Data Types
#data_types = {
#    'astropy.table.table.Table': astrotoyz.data_types.astropy_table
#}

# Source Types
src_types = {
    #'Catalog': astrotoyz.catalog.Catalog
    'AstropyTable': astrotoyz.data_types.astropy_table.AstropyTable
}


import numpy as np


def primes(imax):
    """
    Returns prime numbers up to imax.

    Parameters
    ----------
    imax: int
        The number of primes to return. This should be less or equal to 10000.

    Returns
    -------
    result: list
        The list of prime numbers.
    """

    p = range(10000)
    result = []
    k = 0
    n = 2

    if imax > 10000:
        raise ValueError("imax should be <= 10000")

    while len(result) < imax:
        i = 0
        while i < k and n % p[i] != 0:
            i = i + 1
        if i == k:
            p[k] = n
            k = k + 1
            result.append(n)
            if k > 10000:
                break
        n = n + 1

    return result


def do_primes(n, usecython=False):
    if usecython:
        from .example_c import primes as cprimes
        print('Using cython-based primes')
        return cprimes(n)
    else:
        print('Using pure python primes')
        return primes(n)


def main(args=None):

    from astropy.utils.compat import argparse
    from time import time

    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('-c', '--use-cython', dest='cy', action='store_true',
                        help='Use the Cython-based Prime number generator.')
    parser.add_argument('-t', '--timing', dest='time', action='store_true',
                        help='Time the Fibonacci generator.')
    parser.add_argument('-p', '--print', dest='prnt', action='store_true',
                        help='Print all of the Prime numbers.')
    parser.add_argument('n', metavar='N', type=int,
                        help='Get Prime numbers up to this number.')

    res = parser.parse_args(args)

    pre = time()
    primes = do_primes(res.n, res.cy)
    post = time()

    print('Found {0} prime numbers'.format(len(primes)))
    print('Largest prime: {0}'.format(primes[-1]))

    if res.time:
        print('Running time: {0} s'.format(post - pre))

    if res.prnt:
        print('Primes: {0}'.format(primes))



# Core classes and functions for Astro-Toyz
# Copyright 2015 by Fred Moolekamp
# License: LGPLv3
from toyz.utils.errors import ToyzError

class AstroToyzError(ToyzError):
    """
    Class for errors initiating in Astro Toyz
    """
    pass


# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import

from distutils.core import Extension
from os.path import join

def get_package_data():
    return {
        'astrotoyz': [
            'static/astro.css',
            'static/catalog.js',
            'static/sextractor.js',
            'static/spectrum.js',
            'static/astro.js',
            'static/spectrum.css',
            'static/viewer.js',
            'templates/*.*',
            'static/icons/*.*'],
        }


def requires_2to3():
    return False


"""
Tools for Astro-Toyz Viewer
"""
# Copyright 2015 by Fred Moolekamp
# License: LGPLv3
from __future__ import print_function, division

import toyz.web.viewer
from toyz.web import session_vars
import astropy.wcs as pywcs
import astropy.io.fits as pyfits
import numpy as np
import astrotoyz as astro
import astrotoyz.core

def get_wcs(file_info, hdulist):
    """
    Load world coordinates for the current FITS image
    """
    set_wcs_info = False
    if not hasattr(session_vars, 'wcs_info'):
        session_vars.wcs_info = {}
        set_wcs_info = True
    elif (file_info['filepath'] != session_vars.wcs_info['filepath'] or
            int(file_info['frame']) != session_vars.wcs_info['frame']):
        set_wcs_info = True
    if set_wcs_info==True:
        frame = int(file_info['frame'])
        session_vars.wcs_info['filepath'] = file_info['filepath']
        session_vars.wcs_info['frame'] = frame
        try:
            session_vars.wcs_info['wcs'] = pywcs.WCS(hdulist[frame].header)
        except pywcs.InvalidTransformError:
            # Sometimes PV keywords are not erased from the header when it
            # is converted from TPV to TAN, so if the CTYPE is TAN we
            # remove the useless PV cards from the header to get a 
            # valid WCS
            header = hdulist[frame].header
            if 'TAN' in header['CTYPE1']:
                bad_cards = [card for card in header if card.startswith('PV')]
                for card in bad_cards:
                    del header[card]
                session_vars.wcs_info['wcs'] = pywcs.WCS(header)
            else:
                session_vars.wcs_info = {}
    return session_vars.wcs_info['wcs']

def get_img_data(data_type, file_info, img_info, **kwargs):
    """
    Get data from an image or FITS file
    """
    # First try to fit the data, since the get_2d_fit will center the image at 
    # mean position of the fit
    response = {}
    if 'fit_type' in kwargs:
        fit_response = get_2d_fit(
            file_info, kwargs['fit_type'], kwargs['x'], kwargs['y'], 
            kwargs['width'], kwargs['height'])
        if fit_response['status']=='success':
            kwargs['x'] = int(round(fit_response['fit']['x']))
            kwargs['y'] = int(round(fit_response['fit']['y']))
    # Call regular toyz viewer function to get default information
    response = toyz.web.viewer.get_img_data(data_type, file_info, img_info, **kwargs)
    if 'fit_type' in kwargs and fit_response['status']=='success':
        response.update(fit_response['fit'])
    
    # Now add WCS info
    if file_info['ext']=='fits':
        hdulist = toyz.web.viewer.get_file(file_info)
        data = hdulist[int(img_info['frame'])].data
        wcs = get_wcs(file_info, hdulist)
        if data_type == 'datapoint':
            if (kwargs['x']<data.shape[1] and kwargs['y']<data.shape[0] and
                    kwargs['x']>=0 and kwargs['y']>=0):
                if wcs!=None:
                    wcs_array = wcs.all_pix2world(np.array([[kwargs['x'],kwargs['y']]]),1)
                    response['ra'] = wcs_array[0][0]
                    response['dec'] = wcs_array[0][1]
    return response

def get_img_info(file_info, img_info, **kwargs):
    if file_info['ext']=='fits':
        hdulist = toyz.web.viewer.get_file(file_info)
        hdu = hdulist[int(file_info['frame'])]
        try:
            coord_ranges=[map(int,coord_range.split(':')) for 
                coord_range in hdu.header['DETSEC'][1:-1].split(',')]
        except KeyError:
            coord_ranges=[[1,hdu.data.shape[1]],[1,hdu.data.shape[0]]]
        img_info['coord_range'] = {
            'x': coord_ranges[0],
            'y': coord_ranges[1]
        }
    return img_info

def get_2d_fit(file_info, fit_type, x, y, width, height, **kwargs):
    hdulist = toyz.web.viewer.get_file(file_info)
    wcs = get_wcs(file_info, hdulist)
    hdu = hdulist[int(file_info['frame'])]
    
    # Load user specified tile
    dx = width>>1
    dy = height>>1
    xmin = max(0, x-dx)
    ymin = max(0, y-dy)
    xmax = min(hdu.data.shape[1], x+dx)
    ymax = min(hdu.data.shape[0], y+dy)
    init_data = hdu.data[ymin:ymax,xmin:xmax]
    # Center the tile on the pixel with the highest value
    y_center,x_center = np.unravel_index(init_data.argmax(),init_data.shape)
    xmin = max(0, xmin+x_center-dx)
    ymin = max(0, ymin+y_center-dy)
    xmax = min(hdu.data.shape[1], xmin+width+1)
    ymax = min(hdu.data.shape[0], ymin+height+1)
    data = hdu.data[ymin:ymax,xmin:xmax]
    
    fit, pcov = astro.detect_sources.fit_types[fit_type](data)
    param_map = astro.detect_sources.fit_dtypes[fit_type]
    
    if len(param_map)!=len(fit):
        raise astrotoyz.core.AstroToyzError("Fit parameters did not match parameter map")
    if len(fit)>0:
        params = {value[0]:fit[n] for n, value in enumerate(param_map)}
        params['x'] = xmin+params['x']
        params['y'] = ymin+params['y']
        params['coords'] = str(params['x'])+', '+str(params['y'])
        if wcs!=None:
            wcs_array = wcs.all_pix2world(np.array([[params['x'], params['y']]]),1)
            params['ra'] = wcs_array[0][0]
            params['dec'] = wcs_array[0][1]
        response = {
            'id': 'get_2d_fit',
            'status': 'success',
            'fit': params
        }
    else:
        response = {
            'id': 'get_2d_fit',
            'status': 'fit failed',
            'fit': {}
        }
    return response


from __future__ import division,print_function

import sys
import numpy as np
import numpy.lib.recfunctions as rfn
import scipy.ndimage.filters as filters
import scipy.ndimage as ndimage
from scipy.optimize import curve_fit
import multiprocessing

from toyz.web import session_vars
import toyz.utils.core as core
import astrotoyz.core

# Format of the output for each fit type
fit_dtypes={
    'circular_moffat':[
        ('amplitude',float),
        ('x',float),
        ('y',float),
        ('fwhm',float),
        ('beta',float),
        ('floor',float)
    ],
    'elliptical_moffat':[
        ('amplitude',float),
        ('x',float),
        ('y',float),
        ('fwhm1',float),
        ('fwhm2',float),
        ('beta',float),
        ('angle',float),
        ('floor',float)
    ],
    'fast':[
        ('amplitude',float),
        ('x',float),
        ('y',float),
        ('fwhm1',float),
        ('fwhm2',float),
        ('beta',float),
        ('angle',float),
        ('floor',float),
        ('status', float)
    ],
    'circular_gaussian':[
        ('amplitude',float),
        ('x',float),
        ('y',float),
        ('std_dev',float),
        ('floor',float),
    ],
    'elliptical_gaussian':[
        ('amplitude',float),
        ('x',float),
        ('y',float),
        ('std_x',float),
        ('std_y',float),
        ('angle',float),
        ('floor',float)
    ],
    'no_fit':[
        ('x',float),
        ('y',float)
    ]
}
fit_columns = {fit: [v[0] for v in value] for fit,value in fit_dtypes.items()}

def get_circle_foot(radius):
    """
    get_circle_foot
    
    Generates a circular binary structure with a given radius in O(n) time
    
    Parameters
    ----------
    radius: int
        - radius of the binary structure
    
    Returns
    -------
    footprint: 2d numpy array
        - Circular binary structure with 1's inside the circle and 0's outside
    """
    footprint=np.zeros((2*radius+1,2*radius+1),dtype=int)
    footprint_length=len(footprint)
    for n,row in enumerate(footprint):
        y=abs(radius-n)
        if n>0 and n<footprint_length-1:
            xmin=radius-round(np.sqrt(radius**2-y**2))
        else:
            delta=y
            while round(np.sqrt(delta**2+y**2))>radius:
                delta-=1
            xmin=y-delta
        xmax=footprint_length-xmin
        row[xmin:xmax]=1
    return footprint

def detect_sources(img_data,threshold,aperture_type='radius',size=5,footprint=None,
                    bin_struct=None, sigma=2,saturate=None,margin=None):
    """
    Erodes the background to isolate sources and selects the maximum as approximate positions of sources

    Parameters
    ----------
    img_data: numpy 2D array
        image data
    
    threshold: float
        minimum pixel value above the background noise
    
    size: int, optional
        width of the area in which to search for a maximum (for each point)
    footprint: numpy 2D array (dtype=boolean),optional
        Instead of supplying a size, a footprint can be given of a different shape to use for finding a footprint
            example:
                footprint=np.array([
                    [0,0,1,0,0],
                    [0,1,1,1,0],
                    [1,1,1,1,1],
                    [0,1,1,1,0],
                    [0,0,1,0,0]
                ])
            The above example would only seach for a maximum in the pixels labeled by 1 in the region
            centered on a given pixel
    Note: either a size or a footprint must be secified
    
    bin_struct: 2D numpy array,optional
        Minimum structure that regions of the image are shrunk down to in order to isolate maxima
    sigma: float, optional
        This function uses a guassian filter to smooth the image (only for detection of sources),
        which helps eliminate multiple maximum detections for the same object. 'sigma' describes the
        standard deviation of the gaussian kernel used in the filter
    saturate: float,optional
        Value at which CCD's for the detector become saturated and are no longer linear
    margin: int, optional
        Sources close to the edges can be cut off to prevent partial data from becoming mixed up with good detections
      
    Returns
    -------  
    maxima: numpy 2D array
        Approximate locations of the source maximum values.
        To get more accurate positions each maxima should be fit to a desired profile
    """
    # Make a mask where elements above the threshold are True and below the threshold are False.
    # This essentially removes the background and leaves islands of 1's, representing possible sources
    binData=img_data>=threshold

    # The binary_opening function shrinks all of the 'islands' from the previous step into binary structes,
    # then it dilates them again back to their original shape and width.
    # Shape of the created binary structure (if not specified by the user):
    #   010
    #   111
    #   010
    
    if bin_struct is None:
        bin_struct=ndimage.generate_binary_structure(2,1)
    binData=ndimage.binary_opening(binData,structure=bin_struct)

    # Use our binary data to mask the image and blur the image so get rid of small local maxima that will
    # give us false positive sources
    data=filters.gaussian_filter(binData*img_data,sigma=sigma)

    # The maximum/minimum filters select max/min value in a square with sides length 'size' centered on
    # each element. Filter out all of the objects below the threshold
    params={'input':data}
    if aperture_type=='width':
        params['size']=size
    elif aperture_type=='radius':
        params['footprint']=get_circle_foot(size)
    elif aperture_type=='footprint':
        params['footprint']=footprint
    else:
        raise astrotoyz.core.AstroToyzError('Invalid aperture type in detect_sources')        
    
    # Search for the maximum and minimum points to determine the amplitude of the pixel above its neighboring pixels
    # Note: this only gives the amplitude above the background if the background is within size/2 (or the footprint)
    # of a given pixel.
    data_max=filters.maximum_filter(**params)
    maxima=(data==data_max)
    data_min=filters.minimum_filter(**params)
    diff=((data_max-data_min)>threshold)
    maxima[diff==0]=0

    # Filter out the saturated objects
    if saturate is not None:
        maxima[data>saturate]=0
    
    # Remove the sources near the margins that will be cut off.
    # TODO: Dump these in another file as they will still be useful in determining isolated
    # neighbors for PSF stars
    if margin is None:
        margin=int(size/2)
    if isinstance(margin,list):
        maxima[-margin[0]:,:]=0
        maxima[:margin[1],:]=0
        maxima[:,-margin[2]:]=0
        maxima[:,:margin[3]]=0
    else:
	    maxima[-margin:,:]=0
	    maxima[:margin,:]=0
	    maxima[:,-margin:]=0
	    maxima[:,:margin]=0
    lbl,nbrLbl=ndimage.label(maxima)
    return maxima

def circular_moffat((x,y),amplitude,x_mean, y_mean,beta,alpha,floor):
    """
    Uses 2d array of data to calculate a moffat distribution at the point (x,y), then flattens the data
    into a 1d array for processing.
    
    Parameters
    ----------
    (x,y): tuple of 2D numpy arrays
        Grid of points used to calculate the value of the function
        example:x=np.linspace(0,width-1,width)
                y=np.linspace(0,amplitude-1,amplitude)
                x,y=np.meshgrid(x, y)
    x_mean,y_mean: floats
        location of the maximum (or peak) of the distribution
    alpha,beta: floats
        parameters that describe the shape of a moffat distribution
    floor: float
        lowest asymptotic point of the distribution
    amplitude: float
        maximum amplitude of the distribution above the floor
    
    Returns
    -------
    moff.ravel(): 1D numpy array
        Value of the distribution for the given set of parameters, flattened into a 1D array
        so that it can be used with the curve_fit function
    """
    moff=floor + amplitude/((1+(((x-x_mean)**2+(y-y_mean)**2)/alpha**2))**beta)
    return moff.ravel()

def fit_circular_moffat(data,init_params={}):
    """
    Fits a 2d numpy array to a symmetric Moffat distribution
    
    Parameters
    ----------
    data: 2D numpy array
        image data
    
    Returns
    -------
    fit_result: 2D numpy array
        list of best fit parameters in a form given by src_dtypes['circular moffat']
    pcov: 2D numpy array
        Covariant matrix that describes the error in the fit (but in an 'unscientific' way).
        This needs to be improved to get accurate error estimates
    """
    x = np.linspace(0, data.shape[1]-1, data.shape[1])
    y = np.linspace(0, data.shape[0]-1, data.shape[0])
    x, y = np.meshgrid(x, y)
    
    # Guess initial parameters
    floor = np.median(data)
    amplitude=data.max()-floor
    x_mean=data.shape[1]/2
    y_mean=data.shape[0]/2
    fwhm=np.sqrt(np.sum((data>floor+amplitude/2.).flatten()))
    beta=3.5
    alpha = 0.5*fwhm/np.sqrt(2.**(1./beta)-1.)
    initial_guess=(amplitude,x_mean,y_mean,beta,alpha,floor)
    
    # Attempt fit and return empty lists if it does not converge
    try:
        fit_result,pcov=curve_fit(circular_moffat,(x,y),data.ravel(),p0=initial_guess)
    except RuntimeError:
        return [],[]
    # Convert alpha into a FWHM
    fit_result[3]=np.sqrt(2.**(1./beta)-1.)*fit_result[3]*2
    return fit_result,pcov

def elliptical_moffat((x,y),amplitude,x_mean,y_mean,alpha1,alpha2,beta,angle,floor):
    """
    Uses 2d array of data to calculate a moffat distribution at the point (x,y), then flattens the data
    into a 1d array for processing.
    
    Parameters
    ----------
    (x,y): tuple of 2D numpy arrays
        Grid of points used to calculate the value of the function
        example:x=np.linspace(0,width-1,width)
                y=np.linspace(0,amplitude-1,amplitude)
                x,y=np.meshgrid(x, y)
    x_mean,y_mean: floats
        location of the maximum (or peak) of the distribution
    alpha1,alpha2,beta: floats
        parameters that describe the shape of a moffat distribution
    angle: float
        describes the rotation angle
    floor: float
        lowest asymptotic point of the distribution
    amplitude: float
        maximum amplitude of the distribution above the floor
    
    Returns
    -------
    moff.ravel(): 1D numpy array
        Value of the distribution for the given set of parameters, flattened into a 1D array
        so that it can be used with the curve_fit function
    """
    A = (np.cos(angle)/alpha1)**2. + (np.sin(angle)/alpha2)**2.
    B = (np.sin(angle)/alpha1)**2. + (np.cos(angle)/alpha2)**2.
    C = 2.0*np.sin(angle)*np.cos(angle)*(1./alpha1**2. - 1./alpha2**2.)
    moff=floor + amplitude/((1.+ A*((x-x_mean)**2) + B*((y-y_mean)**2) + C*(x-x_mean)*(y-y_mean))**beta)
    return moff.ravel()

def fit_elliptical_moffat(data):
    """
    Fits a 2d numpy array to a symmetric Moffat distribution
    
    Parameters
    ----------
    data: 2D numpy array
        image data
    
    Returns
    -------
    fit_result: 2D numpy array
        list of best fit parameters in a form given by src_dtypes['elliptical moffat']
    pcov: 2D numpy array
        Covariant matrix that describes the error in the fit (but in an 'unscientific' way).
        This needs to be improved to get accurate error estimates
    """
    x = np.linspace(0, data.shape[1]-1, data.shape[1])
    y = np.linspace(0, data.shape[0]-1, data.shape[0])
    x, y = np.meshgrid(x, y)
    
    # Generate initial guess
    floor = np.median(data)
    amplitude=data.max()-floor
    x_mean=data.shape[1]/2
    y_mean=data.shape[0]/2
    fwhm=np.sqrt(np.sum((data>floor+amplitude/2.).flatten()))
    beta=3.5
    alpha1 = 0.5*fwhm/np.sqrt(2.**(1./beta)-1.)
    alpha2 = 0.5*fwhm/np.sqrt(2.**(1./beta)-1.)
    angle=0
    initial_guess=(amplitude,x_mean,y_mean,alpha1,alpha2,beta,angle,floor)
    
    # Attempt fit and return empty lists if it does not converge
    try:
        fit_result,pcov=curve_fit(elliptical_moffat,(x,y),data.ravel(),p0=initial_guess)
    except RuntimeError:
        # Fit did not converge
        return [],[]
    
    # Convert alpha 1 and 2 into FWHM measurements
    fit_result[3]=np.sqrt(2.**(1./beta)-1.)*fit_result[3]*2
    fit_result[4]=np.sqrt(2.**(1./beta)-1.)*fit_result[4]*2
    return fit_result,pcov

def circular_gaussian((x,y), amplitude, x_mean, y_mean, std_dev, floor):
    gaussian = floor+amplitude*np.exp(-((x-x_mean)**2+(y-y_mean)**2)/(2*std_dev**2))
    return gaussian.ravel()

def fit_circular_gaussian(data):
    x = np.linspace(0, data.shape[1]-1, data.shape[1])
    y = np.linspace(0, data.shape[0]-1, data.shape[0])
    x, y = np.meshgrid(x, y)
    # Guess initial parameters
    floor = np.median(data)
    amplitude=data.max()-floor
    x_mean=data.shape[1]/2
    y_mean=data.shape[0]/2
    mean = data.mean()
    std_dev = 1
    initial_guess=(amplitude,x_mean,y_mean,std_dev,floor)
    
    # Attempt fit and return empty lists if it does not converge
    try:
        fit_result,pcov=curve_fit(circular_gaussian,(x,y),data.ravel(),p0=initial_guess)
    except RuntimeError:
        return [],[]
    # Convert alpha into a FWHM
    return fit_result,pcov

def elliptical_gaussian((x,y), amplitude, x_mean, y_mean, std_x, std_y, theta, floor):
    a = .5*(np.cos(theta)/std_x)**2 + .5*(np.sin(theta)/std_y)**2
    b = -np.sin(2*theta)/(4*std_x**2) + np.sin(2*theta)/(4*std_x**2)
    c = .5*(np.sin(theta)/std_x)**2 + .5*(np.cos(theta)/std_y)**2
    exp = np.exp(-(a*(x-x_mean)**2 + 2*b*(x-x_mean)*(y-y_mean) + c*(y-y_mean)**2))
    gaussian = floor+amplitude*exp
    return gaussian.ravel()

def fit_elliptical_gaussian(data):
    x = np.linspace(0, data.shape[1]-1, data.shape[1])
    y = np.linspace(0, data.shape[0]-1, data.shape[0])
    x, y = np.meshgrid(x, y)
    # Guess initial parameters
    floor = np.median(data)
    amplitude = data.max()-floor
    x_mean = data.shape[1]/2
    y_mean = data.shape[0]/2
    mean = data.mean()
    std_x = 1
    std_y = 1
    theta = 0
    initial_guess=(amplitude, x_mean, y_mean, std_x, std_y, theta, 0)
    # Attempt fit and return empty lists if it does not converge
    try:
        fit_result,pcov=curve_fit(elliptical_gaussian,(x,y),data.ravel(),p0=initial_guess)
    except RuntimeError:
        return [],[]
    # Convert alpha into a FWHM
    return fit_result,pcov

def get_centroid(data):
    """
    Calculate the center using a weighted average for each point
    """
    x = np.linspace(0, data.shape[1]-1, data.shape[1])
    y = np.linspace(0, data.shape[0]-1, data.shape[0])
    x, y = np.meshgrid(x, y)
    x = np.average(x, weights=data)
    y = np.average(y, weights=data)
    coords = [x,y]
    return coords, []

# Map fit types to function defined above
fit_types={
    'circular_moffat': fit_circular_moffat,
    'elliptical_moffat': fit_elliptical_moffat,
    'circular_gaussian': fit_circular_gaussian,
    'elliptical_gaussian': fit_elliptical_gaussian,
    'no_fit': get_centroid
}

# Multiprocessing base on PyMOTW by Doug Hellmann:
# http://pymotw.com/2/multiprocessing/communication.html
class FitWorker(multiprocessing.Process):
    def __init__(self, task_queue, result_queue, data, radius, fit_method):
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue
        self.data = data
        self.radius = radius
        self.fit_method = fit_method
    
    def run(self):
        print('running',self.name)
        fit_func=fit_types[self.fit_method]
        columns = fit_columns[self.fit_method]
        radius = self.radius
        data = self.data
        while True:
            params = self.task_queue.get()
            try:
                if params is not None:
                    x = params['x']
                    y = params['y']
                    xmin=max(x-radius,0)
                    xmax=min(x+radius+1,data.shape[1])
                    ymin=max(y-radius,0)
                    ymax=min(y+radius+1,data.shape[0])
                    best_fit,pcov=fit_func(data[ymin:ymax,xmin:xmax])
                    if len(best_fit)==0:
                        self.task_queue.task_done()
                        self.result_queue.put(tuple([np.nan for i in range(len(columns))]))
                    else:
                        best_fit[1] += x
                        best_fit[2] += y
                        self.task_queue.task_done()
                        self.result_queue.put(tuple(best_fit))
                else:
                    print(self.name,'received exit')
                    self.task_queue.task_done()
                    break
            except Exception as e:
                import traceback
                print('exception in fitting:')
                print(traceback.format_exc())
                print('\n\n\n')
                self.task_queue.task_done()
                self.result_queue.put(tuple([np.nan for i in range(len(columns))]))
        print(self.name,'finished')
        return

def find_stars(img_data, aperture_type='radius', maxima_size=5, 
        maxima_sigma=2, maxima_footprint=None, aperture_radii=[], threshold=None,
        saturate=None, margin=None, bin_struct=None, fit_method='elliptical moffat',
        wcs=None):
    """
    Detect possible sources in an image and attempt to fit them to a specified profile.
    
    Parameters
    ----------
    img_data: 2D numpy array
        Image data
    aperture_type: string
        Type of aperture to use when searching for local maxima. The options are:
            'width': square with width specified by maxima_size
            'radius': circule with radius specified by maxima_size
            'footprint': binary structure with 1's representing 
    maxima_size: int,optional
        Either width of the area or the radius of a circle in which to search for a maximum (for each point)
    maxima_footprint: numpy 2D array (dtype=boolean),optional
        Instead of supplying a size, a footprint can be given of a different shape to use for finding a footprint
            example:
                footprint=np.array([
                    [0,0,1,0,0],
                    [0,1,1,1,0],
                    [1,1,1,1,1],
                    [0,1,1,1,0],
                    [0,0,1,0,0]
                ])
            The above example would only seach for a maximum in the pixels labeled by 1 in the region
            centered on a given pixel
    aperture_radii: list,optional
        List of radii to use to fit the source. In general this should be 5 times the fwhm of the source.
    threshold: float,optional
        Minimum pixel value above the background noise
    saturate: float, optional
        Value at which CCD's for the detector become saturated and are no longer linear
    margin: int, optional
        Sources close to the edges can be cut off to prevent partial data from becoming mixed up with good detections
    bin_struct: 2d numpy array, optional
        Minimum structure that regions of the image are shrunk down to in order to isolate maxima
    fit_method: str
        Type of fit to use to get centroid positions and approximate photometric parameters.
        This step can be skipped by choosing fit_method='no fit'.
    
    Returns
    -------
    best_fits: numpy structured array
        Structured array based on the fit method chosen (given by the fit_dtypes dict).
    no_fit: numpy structured array
        x and y coordinates of sources that could not be fit
    """
    #core.progress_log('Searching for point sources...')
    # Estimate the background by assuming that the middle 80% of the pixels in the 
    # image are background
    if threshold is None:
        sorted_data=np.sort(img_data.flatten())
        back_min_idx=int(sorted_data.size*0.1)
        back_max_idx=int(sorted_data.size*0.9)
        back_estimate=sorted_data[back_min_idx:back_max_idx]
        back_mean=np.mean(back_estimate)
        back_median=np.median(back_estimate)
        back_std=np.std(back_estimate)
        back_min=back_estimate[0]
        back_max=back_estimate[-1]
        threshold=max(abs(back_mean-back_min),abs(back_max-back_mean))
        
        info='\n'.join([
            'Backround estimate minimum:'+str(back_min),
            'Background estimate maximum:'+str(back_max),
            'median:'+str(back_median),
            'mean:'+str(back_mean),
            'standard deviation'+str(back_std),
            'threshold:'+str(threshold)
        ])
        #core.progress_log(info)
    # Find all the point sources and their approximate positions
    sources=detect_sources(img_data,threshold,aperture_type,maxima_size,
        maxima_footprint,bin_struct,maxima_sigma,saturate,margin)
    src_indices=np.where(sources)
    #core.progress_log('Number of stars: '+str(src_indices[0].size))
    
    # Fit the sources to a valid fit method. 
    if fit_method not in fit_types.keys():
        raise astrotoyz.core.AstroToyzError(
            "Invalid fit method, please choose from '"+"','".join(fit_types))
    #core.progress_log('Fitting points')
    fit_func=fit_types[fit_method]
    step=0
    if len(aperture_radii)==0:
        radius=int(maxima_size*3/4)
    else:
        radius=aperture_radii[0]
    
    tasks = multiprocessing.JoinableQueue()
    results = multiprocessing.Queue()

    # Start processes
    num_processes = multiprocessing.cpu_count()
    print('Creating {0} processes'.format(num_processes))
    processes = [FitWorker(tasks, results, img_data, radius, fit_method) 
                    for i in xrange(num_processes)]
    for p in processes:
        p.start()
    num_sources = len(src_indices[0])
    # TODO: use this to test detect sources: 
    num_sources = 5
    for i in range(num_sources):
        x=src_indices[1][i]
        y=src_indices[0][i]
        tasks.put({'x': x,'y': y})
    
    # Add a poison pill for each process
    for p in range(num_processes):
        tasks.put(None)
    # Wait for all of the tasks to finish
    tasks.join()
    
    # Initialize the array to save computation time
    # and initialize to NaN in case of any bad rows
    sources = np.zeros(shape=(num_sources,), dtype=fit_dtypes[fit_method])
    sources.fill(np.nan)
    # get the results
    print('num sources', num_sources)
    for i in xrange(num_sources):
        print('i', i)
        result = results.get()
        #print('i', i, result)
        sources[i] = result
    return sources


"""
Data Types for astrotoyz
"""
from __future__ import division,print_function
from toyz.utils.sources import DataSource
from toyz.utils.errors import ToyzError

class AstroDataError(ToyzError):
    pass

class AstropyTable(DataSource):
    def __init__(self, *args, **kwargs):
        DataSource.__init__(self, *args, **kwargs)
    def name_columns(self, columns=None):
        print('entered name columns!!!!!!!!!!!!!!!!!!!!!!')
        if columns is not None:
            table_cols = self.data.colnames
            if len(columns) != len(table_cols):
                raise AstroDataError("Number of new columns does not match number of table columns")
            for n,col in enumerate(columns):
                self.data.rename_column(table_cols[n], col)
            self.columns = columns
        else:
            self.columns = self.data.colnames
        return self.data.colnames
    #def rename_column(self, old_column, new_column):
    #    if old_column not in self.data.col_names:
    #        raise AstroDataError("Column name not found in data")
    #    self.data.rename_column(old_column, new_column)
    def check_instance(self, data, data_kwargs={}):
        from astropy.table import Table
        if isinstance(data, Table):
            self.data = Table(data, **data_kwargs)
            self.data_type = 'astropy.table.table.Table'
            return True
        return False
    #def set_data(self, data, data_type, data_kwargs={}):
    #    from astropy.table import Table
    #    self.data = Table(data, **data_kwargs)
    #    self.data_type = 'astropy.table.table.Table'
    def to_dict(self, columns=None):
        import numpy as np
        def isnan(x):
            if np.isnan(x):
                return 'NaN'
            else:
                return x
        if columns is None:
            columns = self.columns
        data_dict = {col: map(isnan, np.array(self.data[col]).tolist()) for col in columns}
        return data_dict
    def remove_rows(self, points):
        self.data.remove_rows(points)


"""
Catalog Tools for Astro-Toyz.
For now, all catalogs are stored in a sqlite database in the web app's db directory.
In the future a more robust I/O module will be implemented to allow users
more flexibility in reading and saving their object catalogs
"""
# Copyright 2015 by Fred Moolekamp
# License: LGPLv3
from __future__ import print_function, division
import os
import pandas
import numpy as np
import json
from datetime import datetime

from toyz.web import session_vars
import toyz.utils.core
import astrotoyz.core

# TODO: Remove this part of the code when a more robust API is created
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship, backref
Base = declarative_base()
from sqlalchemy.orm import sessionmaker

def get_catalog(cid, create='yes', **kwargs):
    """
    Get a catalog loaded into session_vars or try to load a saved catalog.
    
    Parameters
        settings (*dict* ):
            - Dictionary of settings for the catalog
        create (*bool* ):
            - Describes action if the catalog does not exist.
            - If ``create='yes'`` an empty catalog is created using ``settings``
            - If ``create='no'`` a None object is returned
            - If ``create='fail'`` an AstroJobError is raised
    """
    if hasattr(session_vars, 'catalogs'):
        if cid in session_vars.catalogs:
            return session_vars.catalogs[cid]
    else:
        session_vars.catalogs = {}
    # Check the catalogs database for a catalog
    # TODO: change the following lines to:
    # if os.path.exists(os.path.join(settings['file_info']['filepath'])):
    try:
        settings = kwargs['settings']
        filepath = settings['file_info']['filepath']
    except KeyError:
        raise astrotoyz.core.AstroToyzError(
            "You must supply a dict of catalog settings with a filepath to "+
            "load or create a catalog")
    catalog = None
    if filepath!='':
        engine = create_engine(filepath) # connects to the db
        Base.metadata.bind = engine # binds the metadata to the engine
        if cid in Base.metadata.tables.keys():
            catalog = load_catalog(settings['file_info'])
    # If the catalog couldn't be loaded, use 'create' to determine what to do
    if catalog is None:
        if create=='yes':
            #init_catalog_db(settings['file_info']['filepath'])
            catalog = Catalog(cid=cid, **kwargs)
        elif create=='no':
            catalog = None
        elif create=='fail':
            raise astrotoyz.core.AstroToyzError("Catalog could not be loaded")
        else:
            raise astrotoyz.core.AstroToyzError("Unrecognized 'create' value '{0}'".format(create))
    return catalog    

def update_catalog(cid, settings, changes):
    """
    If a catalog is saved in session_vars, update it with any changes from the client.
    If the catalog has not been saved in session_vars, the catalog is saved there
    """
    catalog = get_catalog(cid, settings, 'yes')
    if len(changes)>0:
        for change in changes:
            if change['action'] == 'add_src':
                catalog.add_src(change['info'])
            elif change['action'] == 'delete_src':
                catalog.delete_src(change['info'])
            else:
                ToyzJobError("Change action '{0}' is not supported".format(change['action']))
    #print('catalog after updates:', catalog)

def build_src_info(catalog, src_info, file_info):
    """
    Default function to build a catalog id by combining the ``ra`` and ``dec`` fields
    (if available) or ``x`` and ``y`` fields into a coordinate pair
    """
    id_name = catalog.settings['data']['id_name']
    ra_name = catalog.settings['data']['ra_name']
    dec_name = catalog.settings['data']['dec_name']
    if ra_name!='ra' and 'ra' in src_info:
        src_info[ra_name] = src_info['ra']
        del src_info['ra']
    if dec_name!='dec' and 'dec' in src_info:
        src_info[dec_name] = src_info['dec']
        del src_info['dec']
    if ra_name in src_info and dec_name in src_info:
        from astropy.coordinates import SkyCoord
        hdulist = toyz.web.viewer.get_file(file_info)
        hdu = hdulist[int(file_info['frame'])]
        coords = SkyCoord(src_info[ra_name], src_info['dec'], unit='deg')
        src_info[id_name] = coords.to_string('hmsdms')
    else:
        src_info[id_name] = "{0:.6f},{0:.6f}".format(src_info['x'], src_info['y'])
    return src_info

def detect_sources(file_info, cid, settings):
    from astrotoyz.detect_sources import find_stars
    import astrotoyz.viewer
    session_vars.catalogs[cid] = None
    hdulist = toyz.web.viewer.get_file(file_info)
    wcs = astrotoyz.viewer.get_wcs(file_info, hdulist)
    hdu = hdulist[int(file_info['frame'])]
    settings['img_data'] = hdu.data
    sources = find_stars(**settings)
    catalog = Catalog(cid, file_info=file_info, data=sources)
    catalog.dropna(inplace=True)
    if wcs is not None:
        from astropy.coordinates import SkyCoord
        id_name = catalog.settings['data']['id_name']
        ra_name = catalog.settings['data']['ra_name']
        dec_name = catalog.settings['data']['dec_name']
        wcs_array = wcs.all_pix2world(catalog['x'], catalog['y'], 1)
        catalog[ra_name] = wcs_array[0]
        catalog[dec_name] = wcs_array[1]
        coords = SkyCoord(ra=wcs_array[0], dec=wcs_array[1], unit='deg')
        catalog[id_name] = coords.to_string('hmsdms')
    else:
        sep = np.zeroes(shape=(catalog.shape[0],),dtype='|S1')
        sep.fill(',')
        new_id = np.core.defchararray.add(catalog['x'].values.astype('|S10'), sep)
        new_id = np.core.defchararray.add(new_id,catalog['y'].values.astype('|S10'))
        catalog[id_name] = new_id
    catalog.set_index(id_name, inplace=True)
    session_vars.catalogs[cid] = catalog;
    print('finished detecting sources')
    return catalog

class CatalogMeta(Base):
    """
    Table containing meta data for all catalog tables in the database
    """
    __tablename__ = 'catalog_meta'
    id = Column(Integer, primary_key=True)
    cid = Column(String, index=True, unique=True)
    name = Column(String)
    settings = Column(String)

class CatalogLog(Base):
    """
    Table containing logs for each catalog in the database
    """
    __tablename__ = 'catalog_logs'
    id = Column(Integer, primary_key=True)
    action = Column(String)
    log = Column(String)
    cid = Column(Integer, ForeignKey('catalog_meta.cid'), index=True)
    catalog_meta = relationship(CatalogMeta, backref=backref('log', uselist=True))

def init_catalog_db(connect_str):
    """
    Use sqlalchemy to connect to the DB specified by ``connect_str``. 
    """
    engine = create_engine(connect_str) # connects to the db
    Base.metadata.bind = engine # binds the metadata to the engine
    Base.metadata.create_all(engine)

def load_catalog(file_info, load_log=False):
    """
    Load a catalog file. For now this exclusively uses an alchemy connection
    to a DB but in the future this will have methods for various file types,
    or at the very least an export function.
    
    The biggest complication is that in addition to the table information there
    is also metadata (which pandas currently does not support) and a log of
    changes to a catalog.
    """
    cid = file_info['file_settings']['table']
    connect_str = file_info['filepath']
    engine = create_engine(connect_str) # connects to the db
    Base.metadata.bind = engine # binds the metadata to the engine
    if cid not in Base.metadata.tables.keys():
        raise astrotoyz.core.AstroToyzError("Catalog not found in database")
    dataframe = pandas.read_sql_table(cid, engine)
    DBSession = sessionmaker(bind=engine)
    session = DBSession()
    meta = session.query(CatalogMeta).filter(CatalogMeta.cid==cid).first()
    if meta is None:
        raise astrotoyz.core.AstroToyzError("Could not find catalog meta data")
    settings = json.loads(meta.settings)
    if load_log:
        log = pandas.read_sql_query(
            "SELECT * FROM log WHERE cid='{0}'".format(cid),
            engine)
    else:
        log = None
    catalog = Catalog(cid, file_info, name=meta.name, log=log, data=dataframe)
    return catalog

class Catalog(pandas.DataFrame):
    """
    Pandas Dataframe with additional methods to store metadata, log information, and
    functions relating to adding/removing point sources
    """
    def __init__(self, cid, file_info, settings={}, name=None, data=None, log=None, **df_settings):
        if data is not None:
            df_settings['data'] = data
        pandas.DataFrame.__init__(self, **df_settings)
        
        # Update Catalog specific attributes
        self.cid = cid
        if name is not None:
            self.name = name
        else:
            if 'name' in settings:
                self.name = settings['name']
                del settings['name']
            else:
                self.name = self.cid
        # Store settings
        default_settings = {
            'creation': {
                'creation_time': str(datetime.now()),
                'fits_file': file_info,
                'software_version': {
                    'toyz': 'alpha', # TODO: make toyz.version.version work properly
                    'astrotoyz': 'alpha' # TODO: make astrotoyz.version.version work properly
                }
            },
            'file_info': {
                'filepath': '',
                'file_type': '',
                'file_settings': {}
            },
            'data': {
                'fit_method': 'elliptical_moffat',
                'ra_name': 'ra',
                'dec_name': 'dec',
                'id_name': 'id',
                'min_sep': {
                    'wcs': [1,'arcsec'],
                    'px': 5
                },
                'build_src_info': {
                    'module': 'astrotoyz.catalog',
                    'func': 'build_src_info'
                }
            }
        }
        self.settings = toyz.utils.core.merge_dict(default_settings, settings, True)
        if (self.index.name is None and self.index.names[0] is None and 
                self.settings['data']['id_name'] in self.columns):
            self.set_index(self.settings['data']['id_name'], inplace=True)
        if log is not None:
            self.log = log
        else:
            log = pandas.DataFrame(columns=['cid', 'action', 'entry'])
        self.new_log = pandas.DataFrame(columns=['cid', 'action', 'entry'])
    
    def save(self, filepath=None):
        """
        Save the catalog. If a filepath is specified that is different than the current
        filepath, update the catalogs settings.
        """
        # TODO: Change the following code to allow for more genral file types
        if filepath is None:
            connect_str = self.settings['file_info']['filepath']
        elif filepath != self.settings['file_info']['filepath']:
            connect_str = filepath
            self.settings['file_info']['filepath'] = connect_str
        engine = create_engine(connect_str) # connects to the db
        self.to_sql(self.cid, engine, if_exists='replace', index=False)
        
        # Save meta data and new log entries
        DBSession = sessionmaker(bind=engine)
        session = DBSession()
        # Encode settings dict as a json string
        meta_record = session.query(CatalogMeta).filter(CatalogMeta.cid==self.cid)
        if meta_record.first() is None:
            meta = CatalogMeta(cid=self.cid, name=self.name, 
                           settings=json.dumps(self.settings))
            session.add(meta)
        else:
            meta = meta_record
            meta.update({
                'cid': self.cid,
                'name': self.name,
                'settings': json.dumps(self.settings)
            })
        # Save the new log entries, then add them to the (old) log and clear the new log
        self.new_log.to_sql('log', engine, if_exists='append')
        self.log.append(self.new_log)
        self.new_log = pandas.DataFrame(columns=['action', 'cid', 'entry'])
        session.commit()
        Base.metadata.reflect(engine) # adds the table created by to_sql to the metadata
        Base.metadata.create_all(engine) # updates the metadata with the new tables
        return True
    
    def log(self, action, log):
        """
        Log changes to catalog
        """
        self.new_log.loc[self.new_log.shape[0]] = pandas.Series({
            'cid': self.cid,
            'action': action,
            'log': log
        })
    
    def get_log(self, include_new=True):
        """
        Get the full log for the catalog from the log file. If ``include_new`` is ``True``,
        the changes made since the last save are also stored.
        """
        # TODO: make this work for more general log file types
        engine = create_engine(self.settings['file_info']['filepath']) # connects to the db
        Base.metadata.bind = engine # binds the metadata to the engine
        DBSession = sessionmaker(bind=engine)
        session = DBSession()
        log = session.query(CatalogLog).filterby(cid=self.cid)
        log = [{l.action: l.log} for l in log]
        self.log = log
        if include_new:
            log.extend(self.new_log)
        return log    
    
    def get_min_sep(self):
        """
        Get the minimum separation for a point to be considered a new source
        """
        import astropy.units as u
        # Check to make sure that the source isn't already in the catalog
        if 'min_sep' in self.settings['data']:
            if 'wcs' in self.settings['data']['min_sep']:
                minsep = self.settings['data']['min_sep']['wcs']
                sep = minsep[0]*getattr(u,minsep[1])
            else:
                if 'px' in self.settings['data']['min_sep']:
                    sep = self.settings['data']['min_sep']['px']
                else:
                    sep = 1*u.m
        else:
            sep = 1*u.arcsec
        return sep
    
    def get_src_info(self, **kwargs):
        if (('ra' in kwargs and 'dec' in kwargs) or
                (self.settings['data']['ra_name'] in kwargs and 
                self.settings['data']['dec_name'] in kwargs)):
            src_info = kwargs
        else:
            import astrotoyz.viewer
            import copy
            params = copy.deepcopy(kwargs)
            
            if 'fit_type' not in params:
                params['fit_type'] = 'elliptical_gaussian'
            if 'width' not in params:
                params['width'] = 20
            if 'height' not in params:
                params['height'] = 20
            fit = astrotoyz.viewer.get_2d_fit(**params)
            if fit['status'] == 'success':
                src_info = fit['fit']
            else:
                raise astrotoyz.core.AstroToyzError("Fit did not converge")
        return src_info
    
    def select_src(self, src_info, file_info, **kwargs):
        """
        Find the object in the catalog the is the nearest neighbor to the current source.
        If the object is within Catalog.settings['minsep'] (``px`` or ``wcs`` depending
        on the coords given) then it is selected
        """
        kwargs.update(src_info)
        src_info.update(self.get_src_info(file_info=file_info, **kwargs))
        sep = self.get_min_sep()
        idx, d2d, src = self.get_nearest_neighbors(self, coords=src_info)
        if d2d<sep and d2d:
            src = {s: src[n] for n,s in enumerate(src.keys())}
        else:
            src = {}
        return src
    
    def add_src(self, src_info, file_info, **kwargs):
        """
        Add a source to to the current catalog if it hasn't already been added.
        Note: ``min_sep`` describes the minimum distance for a point to be considered
        a new source (as opposed to a rounding error). The current default is 5 pixels or 1 arcsec.
        """
        kwargs.update(src_info)
        src_info.update(self.get_src_info(file_info=file_info, **kwargs))
        sep = self.get_min_sep()
        if self.shape[0]>0:
            idx, d2d, src = self.get_nearest_neighbors(self, coords=src_info)
        
        if self.shape[0]==0 or d2d>sep:
            import importlib
            # Call a function to configure the source for the current catalog
            # (this can be a user specified function specified when the catalog
            # was created)
            build_module = importlib.import_module(
                self.settings['data']['build_src_info']['module'])
            build_func = self.settings['data']['build_src_info']['func']
            src_info = getattr(build_module, build_func)(self, src_info, file_info)
            print('catalog before', self)
            # Bad fix for adding a new source, need to change this
            if self.shape[0]==0:
                for k,v in src_info.items():
                    self[k] = [v]
                print('catalog', self)
            else:
                self.loc[src_info[self.settings['data']['id_name']]] = pandas.Series(src_info)
            self.log('add_src', src_info)
            return src_info
        return {}
    
    def delete_src(self, src_info):
        """
        Delete a source from the catalog.
        """
        id_name = self.settings['data']['id_name']
        row = self.loc[src_info[id_name]]
        self.drop(src_info[id_name], inplace=True)
        # Notify the user if no matching sources were found
        #print('catalog after delete', self)
        if len(row)==0:
            return False
        return True
    
    def get_nearest_neighbors(self, coord1=None, coord2=None, coord_type=None, 
            coord_unit=None, coords=None):
        """
        Calls astrotoyz.catalog.get_nearest_neighbors. Since this method is called from
        the catalog object, it allows the user to only specify a 'coords' dictionary that
        contains either 'ra' and 'dec', 'x' and 'y', or the catalogs 'ra_name' and 'dec_name'
        pairs of coordinates and automatically calculates the other fields necessary.
        """
        from astropy.coordinates import SkyCoord
        if coords is None:
            if coord1 is None or coord2 is None or coord_type is None:
                raise astrotoyz.core.AstroToyzError(
                    "Must supply either 'coords' or 'coord1' and 'coord2' and 'coord_type")
        else:
            if (
                self.settings['data']['ra_name'] in coords and 
                self.settings['data']['dec_name'] in coords
            ):
                coord1 = coords[self.settings['data']['ra_name']]
                coord2 = coords[self.settings['data']['dec_name']]
                coord_type = 'wcs'
            elif ('ra' in src_info and 'dec' in src_info and 
                    (coord_type is None or coord_type=='wcs')):
                coord1 = src_info['ra']
                coord2 = src_info['dec']
                coord_type = 'wcs'
            elif ('x' in src_info and 'y' in src_info and
                    (coord_type is None or coord_type=='px')):
                coord1 = src_info['x']
                coord2 = src_info['y']
                coord_type = 'px'
            else:
                raise astrotoyz.core.AstroToyzError(
                    "Adding a source requires a pair of coordinates from 'ra' and 'dec', " +
                    "'x' and 'y', or the catlogs 'ra' and 'dec' fields")
        if coord_type=='wcs':
            if coord_unit is None:
                coord_unit = 'deg'
            cat1 = self[self.settings['data']['ra_name']]
            cat2 = self[self.settings['data']['dec_name']]
            c1 = SkyCoord(ra=coord1, dec=coord2, unit=coord_unit)
            cat_coords = SkyCoord(ra=cat1, dec=cat2, unit=coord_unit)
        elif coord_type=='px':
            coord_unit = 'm' # a unit of length is needed to use astropy's matching function
            cat1 = self['x'].values
            cat2 = self['y'].values
            c1 = SkyCoord(x=coord1, y=coord2, z=0, unit='m', representation='cartesian')
            cat_coords = SkyCoord(x=cat1, y=cat2, z=0, unit= 'm', representation='cartesian')
        else:
            raise astrotoyz.core.AstroToyzError("Unrecognized coord_type")
        idx, d2d, d3d = c1.match_to_catalog_sky(cat_coords)
        # Since there is no consistent way to deal with numpy scalars, we 
        # take the max of the scalar to get its value (ugh)
        # Good discussion on this at
# http://stackoverflow.com/questions/773030/why-are-0d-arrays-in-numpy-not-considered-scalar
        if idx.shape==():
            idx = idx.max()
        return idx, d2d, self.loc[self.index[idx]]
    def get_markers(self):
        data = self.reset_index()[['id','x','y']].fillna('NaN').values.tolist()
        markers = {
            'columns': ['id','x','y'],
            'data': data
        }
        return markers